# https://pyrocko.org - GPLv3
#
# The Pyrocko Developers, 21st Century
# ---|P------/S----------~Lg----------

import numpy as num

from pyrocko.model.location import Location
from pyrocko.plot.smartplot import Plot
from pyrocko import plot, util
from pyrocko.gui import util as gui_util
from pyrocko import gato


km = 1000.


def time_or_none_to_date_str(t):
    if t is None:
        return '...'
    else:
        return util.time_to_str(t, format='%Y-%m-%d')


class GeometryPlot(Plot):

    def __init__(self, *args, **kwargs):
        self._info = None
        self._array = None
        self._items = []
        self._items_highlight = []
        self._units_factor = 1.0
        self._units = 'm'

        Plot.__init__(
            self,
            ['easting'],
            ['northing'],
            *args,
            **kwargs)

        self.set_aspect('northing', 'easting')
        self.set_units(km, 'km')
        self._connect(
            self.fig.canvas, 'motion_notify_event', self.on_mouse_move)

    def have_sensors(self):
        return self._info and self._info.sensors

    def on_mouse_move(self, event):

        x, y = event.xdata, event.ydata
        if self.have_sensors() and None not in (x, y):

            location = Location(
                lat=self._grid_no_dups.get_effective_origin().lat,
                lon=self._grid_no_dups.get_effective_origin().lon,
                north_shift=y*self._units_factor,
                east_shift=x*self._units_factor,
                depth=0.0)

            self.highlight_closest_sensor(location)
        else:
            self.highlight_closest_sensor(None)

        self.fig.canvas.draw()

    def set_units(self, factor, label):
        self._units_factor = factor
        self._units = label
        self.update_labels()

    def update_labels(self):
        self.set_label('easting', 'Easting [%s]' % self._units)
        self.set_label('northing', 'Northing [%s]' % self._units)

    def set_array(self, array, info):
        self._array = array
        self._info = info
        self._grid_no_dups = None
        self._grid = None
        self._grid_origin = None

        if self.have_sensors():
            self._grid_no_dups = gato.UnstructuredLocationGrid.from_locations(
                info.sensors, ignore_position_duplicates=True)
            self._grid_no_dups.set_origin_to_center()

            self._grid = gato.UnstructuredLocationGrid.from_locations(
                info.sensors)

            self._grid.origin = self._grid_no_dups.get_effective_origin()

        self.update()

    def update(self, *args):
        self.draw()
        self.fig.canvas.draw()

    def get_closest_sensors(self, location):
        if not self.have_sensors():
            return [], num.zeros((0, 3))

        lgrid = gato.UnstructuredLocationGrid.from_locations([location])
        lgrid.origin = self._grid_no_dups.get_effective_origin()
        distances = gato.distances_surface(self._grid, lgrid)[:, 0]
        coords_ned = self._grid.get_nodes('ned')
        order = num.argsort(distances)
        sensors = []
        i = 0
        while i < order.size \
                and num.all(
                    coords_ned[order[0], :] == coords_ned[order[i], :]):
            sensors.append(self._info.sensors[order[i]])
            i += 1

        return sensors, coords_ned[order[:len(sensors)], :]

    def highlight_closest_sensor(self, location):

        axes = self(0, 0)
        while self._items_highlight:
            self._items_highlight.pop().remove()

        win = gui_util.get_app().get_main_window()

        if location is None:
            win.status('')
            return

        if not self.have_sensors:
            return

        sensors, coords_ned = self.get_closest_sensors(location)

        pos = (
            coords_ned[0, 1] / self._units_factor,
            coords_ned[0, 0] / self._units_factor)

        self._items_highlight.extend(
            axes.plot(
                *pos,
                'o',
                color=plot.mpl_color('scarletred1'),
                mec=plot.mpl_color('scarletred3'),
                ms=10))

        codes_list = sorted(set(sensor.codes for sensor in sensors))
        nsl_list = sorted(set(sensor.codes.nsl for sensor in sensors))
        c_list = sorted(set(
            channel.codes.channel
            for sensor in sensors for channel in sensor.channels))
        lat_lon_list = sorted(set(
            sensor.effective_latlon for sensor in sensors))
        elevation_list = sorted(set(
            sensor.elevation for sensor in sensors))
        depth_list = sorted(set(
            sensor.depth for sensor in sensors))

        spans_list = sorted(set('%s - %s' % (
            time_or_none_to_date_str(sensor.tmin),
            time_or_none_to_date_str(sensor.tmax))
            for sensor in sensors))

        label = ', '.join('.'.join(nsl) for nsl in nsl_list)
        self._items_highlight.append(
            axes.annotate(
                label,
                xy=pos,
                xytext=(10, 10),
                textcoords='offset points'))

        message = 'Sensors: %s, Location: %s, Elevation: %s, Depth: %s, ' \
            'Channels: %s, Epochs: %s' % (
                ', '.join(str(codes) for codes in codes_list),
                ', '.join('%g°/%g°' % lat_lon for lat_lon in lat_lon_list),
                ', '.join('%g m' % elevation for elevation in elevation_list),
                ', '.join('%g m' % depth for depth in depth_list),
                ', '.join(c_list),
                ', '.join(spans_list))

        win.status(message)

    def draw(self):
        while self._items:
            self._items.pop().remove()

        array = self._array

        if not self.have_sensors():
            return

        axes = self(0, 0)
        title = array.name + (' (%s)' % array.comment if array.comment else '')
        axes.set_title(title, pad=20)

        coords_ned = self._grid_no_dups.get_nodes('ned')
        vmin = num.min(coords_ned, axis=0)
        vmax = num.max(coords_ned, axis=0)
        mask_eq = vmin == vmax
        vmin[mask_eq] -= 1.0
        vmax[mask_eq] += 1.0

        scale_factor = 1.2
        self.set_lim(
            'northing',
            scale_factor * vmin[0] / self._units_factor,
            scale_factor * vmax[0] / self._units_factor)
        self.set_lim(
            'easting',
            scale_factor * vmin[1] / self._units_factor,
            scale_factor * vmax[1] / self._units_factor)

        self.need_update_layout()

        self._items.extend(
            axes.plot(
                coords_ned[:, 1] / self._units_factor,
                coords_ned[:, 0] / self._units_factor,
                'o',
                color='black'))


__all__ = [
    'GeometryPlot',
]
